@@ -5,27 +5,25 @@
 //! for discrete simulation
 
 //
-// Copyright (c) 1991-2018 Petr Peringer
+// Copyright (c) 1991-2022 Petr Peringer
 //
 // This library is licensed under GNU Library GPL. See the file COPYING.
 //
 
 //
-//  Implementation of interruptable functions (non-preemptive threads/coroutines)
+// Implementation of interruptable functions (non-preemptive, coroutine-like)
 //
-//  This module is the only nonportable module in SIMLIB
-//  We need code to save/restore process stack contents and working setjmp/longjmp
-//  This approach has advantage in small memory requirements.
+// We need code to save/restore process stack contents and working setjmp/longjmp
+// This approach has advantage in smaller memory requirements.
 //
-//  Supported CPU architectures: i386+, x86_64
+// LIMITATIONS:
+//  - do not use (e.g. using pointers) any locals in Process::Behavior() if it
+//    is not in running state (stack contents is moved to heap)
+//  - do not use too big locals in Process::Behavior() for performance reasons
 //
-//  WARNING: dirty hack inside
+// WARNING: dirty hacks inside
 //
 
-// DONE: mark SP, alloc stack [size=runtime-option], change SP, call global function without
-//       params/locals, call Current->Behavior (uses new stack for this)
-//       return: set SP back, ...
-//       Process destructor: free stack
 // TODO: add implementation with stack switching (not copying)
 //       as compile-time option
 // TODO: add implementation using C++20 coroutines?
@@ -41,17 +39,7 @@
 #include <csetjmp>
 #include <cstring>
 
-// basic operating system test
-#if !(defined(__MSDOS__)||defined(__linux__)|| \
-      defined(__WIN32__)||defined(__FreeBSD__))
-# error "module process.cc is not implemented on this operating system"
-#endif
-
-// basic CPU test
-#if !(defined(__i386__)||defined(__x86_64__))
-# error "module process.cc is not ported to this processor architecture"
-#endif
-
+static_assert(sizeof(void*) >= 4, "not tested on <32bit systems");
 
 ////////////////////////////////////////////////////////////////////////////
 // implementation
@@ -61,188 +49,123 @@ namespace simlib3 {
 
 SIMLIB_IMPLEMENTATION;
 
-////////////////////////////////////////////////////////////////////////////
-// Machine dependent macros for direct stack pointer manipulation:
-//
-// GET_STACK_PTR(v)   v = <stack pointer>       (not used)
-// SET_STACK_PTR(v)   <stack pointer> = v
-//
-/////////////////////////////////////////////////////////////////////////////
-#if defined(__BCPLUSPLUS__) || defined(__TURBOC__)  // Borland compilers
-
-// this works on i386+ platforms
-# if defined(__WIN32__)         // 32 bit WINDOWS
-#   define GET_STACK_PTR(v)     { v = (char*)_ESP; }
-#   define SET_STACK_PTR(v)     { _ESP = v; }
-# else                          // 16 bit MSDOS (obsolete, not supported)
-#   if defined(__LARGE__) || defined(__COMPACT__) || defined(__HUGE__)
-#     define GET_STACK_PTR(v)   { v = (char far *)MK_FP(_SS,_SP); }
-#   else
-#     define GET_STACK_PTR(v)   { v = (char*)_SP; }
-#   endif
-#   define SET_STACK_PTR(v)     { _SP = v; }
-# endif
-
-#elif defined(__GNUC__)         // GNU C++ compiler
-
-# if defined(__i386__)          // 32bit: i386+ expected...
-
-#   define GET_STACK_PTR(v)     { asm("movl %%esp,%0":"=r" (v)); }
-#   define SET_STACK_PTR(v)     { asm("movl %0,%%eax": : "m" (v)); \
-                                  asm("movl %eax,%esp");                     \
-                                }
-
-# elif defined(__x86_64__)      // 64bit: Athlon64, ...
-
-#   define GET_STACK_PTR(gvar)  { asm("movq %%rsp,%0": "=r" (gvar)); }
-#   define SET_STACK_PTR(gvar)  { asm("movq %0,%%rsp": : "m" (gvar)); }
-
-# else                          // Other CPU ...
-#   error "process.cc: Unsupported CPU architecture"
-# endif
-
-#else                           // Other compilers not supported
-
-# error "process.cc: Unsupported compiler"
-
-#endif
-/////////////////////////////////////////////////////////////////////////////
-
-
-////////////////////////////////////////////////////////////////////////////
-// machine independent code:
 ////////////////////////////////////////////////////////////////////////////
 
 /**
- * internal structure for storing of process context
+ * internal structure for storing of Process::Behavior() context
  * @ingroup process
  */
 struct P_Context_t {
     jmp_buf status;     //!< stored SP, IP, and other registers
     size_t size;        //!< size of following array (allocated on heap)
-    char stack[1];      //!< stack contents saved
+    char stack[1];      //!< saved stack contents
 };
 
 ////////////////////////////////////////////////////////////////////////////
 // global variables (should be volatile)
+// (P_ means Process)
+// TODO: wrap to thread_local struct together with (almost) all globals
 static jmp_buf P_DispatcherStatusBuffer; //!< setjmp() state before dispatch
 static char *volatile P_StackBase = 0;   //!< global start of stack area
-static char *volatile P_StackBase2 = 0;  //!< for checking start of stack
+static char *volatile P_StackBase2 = 0;  //!< used for checking only
 
 static P_Context_t *volatile P_Context = 0; //!< temporary global process state
 static volatile size_t P_StackSize = 0;     //!< temporary global stack size
 
 ////////////////////////////////////////////////////////////////////////////
-// Support for THREADS implementation debugging:
+// Support for debugging:
 ////////////////////////////////////////////////////////////////////////////
 
 #define EXTRA_DEBUG 0   // 0==off, 1==on
 
 #if EXTRA_DEBUG
 
-/// \def THREAD_DEBUG
+/// \def DEBUG_PROCESS
 /// Macro for detailed debugging of Process switching code
-#define THREAD_DEBUG(n) THREAD_DEBUG_f(n)
+#define DEBUG_PROCESS(n) DEBUG_PROCESS_f(n)
 
-// FIXME
-// We use extra debug print macro, because for 0xC we can not print float
-// format used in DEBUG macro (prints Time using %g) -> segfault
-// [%d format is O.K.] [tested with GCC version 7, 8, 11]
-#   define P_DEBUG(c,f)                         \
-    do{ if( SIMLIB_debug_flag & (c) ) {         \
-          _Print("EXTRA_PROCESS_DEBUG:");       \
-          _Print f; _Print("\n");               \
-    } }while(0)
-
-/// \fn THREAD_DEBUG_f
-/// We need this non-inline function because optimization of volatile global
+/// \fn DEBUG_PROCESS_f
+/// We need? (FIXME) this non-inline function because optimization of volatile global
 /// variable access after longjmp -> setjmp transition in 32bit i686 mode of
 /// GCC 7 and newer versions
-/// TODO: +string table
-static void THREAD_DEBUG_f(int n) __attribute__ ((noinline));
-static void THREAD_DEBUG_f(int n) {
+// TODO: add string table? remove?
+[[gnu::noinline]] static void DEBUG_PROCESS_f(int n) {
     switch(n) {
     case 1:
-        P_DEBUG(DBG_THREAD,("| THREAD_INTERRUPT ***** begin *****"));
-        P_DEBUG(DBG_THREAD,("|   %d) - stack size = %p", n, P_StackSize));
+        DEBUG(DBG_PROCESS,("| PROCESS_INTERRUPT ***** begin *****"));
+        DEBUG(DBG_PROCESS,("|   %d) - stack size = %p", n, P_StackSize));
         break;
     case 2:
         break;
     case 3:
         break;
     case 4:
-        P_DEBUG(DBG_THREAD,("|   %d) THREAD_SAVE_STACK: before setjmp() - context=%p", n, P_Context));
+        DEBUG(DBG_PROCESS,("|   %d) PROCESS_SAVE_STACK: before setjmp() - context=%p", n, P_Context));
         break;
     case 5:
-        P_DEBUG(DBG_THREAD,("|   %d) THREAD_SAVE_STACK: after setjmp()  - context=%p", n, P_Context));
+        DEBUG(DBG_PROCESS,("|   %d) PROCESS_SAVE_STACK: after setjmp()  - context=%p", n, P_Context));
         break;
     case 6:
-        P_DEBUG(DBG_THREAD,("|   %d) THREAD_RESTORE: longjmp() back     - context=%p", n, P_Context));
+        DEBUG(DBG_PROCESS,("|   %d) PROCESS_RESTORE: longjmp() back     - context=%p", n, P_Context));
         break;
     case 7:
-        P_DEBUG(DBG_THREAD,("| THREAD_INTERRUPT ***** end *****"));
+        DEBUG(DBG_PROCESS,("| PROCESS_INTERRUPT ***** end *****"));
         break;
 
     case 0xB:
-        P_DEBUG(DBG_THREAD,("| b) THREAD Shift SP to %p ", P_StackBase2));
+        DEBUG(DBG_PROCESS,("| b) PROCESS Shift SP to %p (or more)", P_StackBase2));
         break;
     case 0xC:
-        P_DEBUG(DBG_THREAD,("| c) THREAD Before longjmp(%p,1)", P_Context->status));
+        DEBUG(DBG_PROCESS,("| c) PROCESS Before longjmp(%p,1)", P_Context->status));
         break;
 
     default:
-        P_DEBUG(DBG_THREAD,("|   %d) - context = %p", n, P_Context));
+        DEBUG(DBG_PROCESS,("|   %d) - context = %p", n, P_Context));
         break;
     }
 }
 #else
-#define THREAD_DEBUG(n)
+#define DEBUG_PROCESS(n)
 #endif
 
 ////////////////////////////////////////////////////////////////////////////
-// special THREADS implementation:
+// Specialized asymmetric stackfull coroutine implementation:
 ////////////////////////////////////////////////////////////////////////////
 
-static void THREAD_INTERRUPT_f() __attribute__ ((noinline)); // special function
+[[gnu::noinline]] static void PROCESS_INTERRUPT_f(); // special function
 
 /// interrupt process behavior execution, continue after return
-#define THREAD_INTERRUPT()                                              \
+#define PROCESS_INTERRUPT()                                              \
 { /* This should be MACRO */                                            \
   /* if(!isCurrent())  SIMLIB_error("Can't interrupt..."); */           \
   this->_status = _INTERRUPTED;                                         \
-  THREAD_INTERRUPT_f();                                                 \
+  PROCESS_INTERRUPT_f();                                                 \
   this->_status = _RUNNING;                                             \
   this->_context = 0;                                                   \
 }
 
 /// does not save context
-#define THREAD_EXIT() \
+#define PROCESS_EXIT() \
     longjmp(P_DispatcherStatusBuffer, 2)  // jump to dispatcher
 
-// TODO: allocation/freeing memory is expensive, should be optimized
+// FIXME: allocation/freeing memory is expensive, should be optimized
 
 /// \def ALLOC_CONTEXT
 /// allocate memory for process context, sz = size of stack area to save
-#define ALLOC_CONTEXT(sz)                                                 \
-    P_Context = (P_Context_t *) new char[sizeof(P_Context_t) + sz];       \
+[[gnu::noinline]] static void ALLOC_CONTEXT(size_t sz) noexcept try
+{
+    P_Context = (P_Context_t *) new char[sizeof(P_Context_t) + sz];
     P_Context->size = sz;
+} catch(std::bad_alloc&) {
+    SIMLIB_error("ALLOC_CONTEXT: no memory");
+}
 
-// bug for new compilers (2018) -- too aggresive optimization of global var
-// access in THREAD_INTERRUPT_f causes longjmp register problem
-#if BUG_IN_32BIT_CODE_USING_GCC_7_plus
-/// free memory of process context
-#define FREE_CONTEXT()                  \
-    delete[] (char *) P_Context;        \
-    P_Context = 0;
-#else
-static void FREE_CONTEXT() __attribute__ ((noinline)); // special function
 /// \fn FREE_CONTEXT
-/// non-inline function for deallocating saved process context
-static void FREE_CONTEXT() {
-    delete[] (char *) P_Context;
+/// deallocate saved process context
+[[gnu::noinline]] static void FREE_CONTEXT() noexcept {
+    delete[] (char *)(P_Context);
     P_Context = 0;
 }
-#endif
 
 ////////////////////////////////////////////////////////////////////////////
 /// Process constructor
@@ -256,7 +179,7 @@ Process::Process(Priority_t p) : Entity(p) {
 
 ////////////////////////////////////////////////////////////////////////////
 /// Process destructor
-/// Sets status to TERMINATED and
+/// Sets state to TERMINATED and
 /// removes process from queue/calendar/waituntil list.
 /// TODO: Warn if this==Current? (e.g. "delete this;" in Behavior())
 Process::~Process()
@@ -266,22 +189,20 @@ Process::~Process()
     //if(this==Current) SIMLIB_warning("Currently running process self-destructed");
 
     // destroy context data
-    delete [] (char*)_context;
+    delete [] static_cast<char*>(_context);
     _context = 0;
 
     _status = _TERMINATED;
 
     if (_wait_until) {
-        //SIMLIB_warning("Process in Wait-Until destructed");
-        _WaitUntilRemove();     // Remove from wait-until list
+        _WaitUntilRemove();     // remove from wait-until list
     }
 
     if (Where() != 0) {         // if waiting in queue
-        //SIMLIB_warning("Process waiting in queue destructed");
         Out();                  // remove from queue, no warning
     }
+
     if (!Idle()) {              // if process is scheduled
-        //SIMLIB_warning("Active process destructed");
         SQS::Get(this);         // remove from calendar
     }
 }
@@ -289,8 +210,7 @@ Process::~Process()
 #if 1
 ////////////////////////////////////////////////////////////////////////////
 /// Name of the process
-/// warning: uses static buffer for generic names
-/// TODO: use std::string
+/// Each process can be named, default is "Process#<number>"
 std::string Process::Name() const
 {
     const std::string name = SimObject::Name();
@@ -305,6 +225,7 @@ std::string Process::Name() const
 /// Interrupt process behavior - this ensures WaitUntil tests
 /// WARNING: use with care - it can run higher (or equal) priority processes
 ///          before continuing
+/// TODO: Works almost like Wait(0), used only with WaitUntil
 void Process::Interrupt()
 {
     Dprintf(("Process#%lu.Interrupt()", _Ident));
@@ -314,7 +235,7 @@ void Process::Interrupt()
     // TODO: use >highest priority to eliminate problem
     Entity::Activate(); // schedule now - can run higher priority
                         //                processes before this one
-    THREAD_INTERRUPT();
+    PROCESS_INTERRUPT();
     // TODO: return to previous priority
 }
 
@@ -323,22 +244,22 @@ void Process::Interrupt()
 void Process::Activate(double t)
 {
     Dprintf(("Process#%lu.Activate(%g)", _Ident, t));
-    Entity::Activate(t);
+    Entity::Activate(t);                        // (re)scheduling
     if (!isCurrent())
         return;
-    // SIMLIB_warning("Process can not activate itself - use Wait");
-    THREAD_INTERRUPT();
+    PROCESS_INTERRUPT();
 }
 
 ////////////////////////////////////////////////////////////////////////////
 /// Wait for dtime
+/// The same as Activate(Time+dtime)
 void Process::Wait(double dtime)
 {
     Dprintf(("Process#%lu.Wait(%g)", _Ident, dtime));
-    Entity::Activate(double (Time) + dtime);    // scheduling
+    Entity::Activate(double (Time) + dtime);    // (re)scheduling
     if (!isCurrent())
         return;
-    THREAD_INTERRUPT();
+    PROCESS_INTERRUPT();
 }
 
 ////////////////////////////////////////////////////////////////////////////
@@ -368,19 +289,19 @@ void Process::Enter(Store & s, unsigned long cap)
 ////////////////////////////////////////////////////////////////////////////
 /// Leave - return cap capacity of store s
 /// and enter first waiting entity from queue, which can use free capacity
-//TODO: should be parametrized: use first-only, first-good, all-good
 void Process::Leave(Store & s, unsigned long cap)
 {
     s.Leave(cap);               // polymorphic interface
+    //TODO: should be parametrized: use first-only, first-usable, all-usable
 }
 
 ////////////////////////////////////////////////////////////////////////////
-/// insert current process into queue
-/// The process can be at most in single queue.
+/// Insert current process into queue
+/// The process can be at most in single queue. Moves if in other queue.
 void Process::Into(Queue & q)
 {
     if (Where() != 0) {
-        SIMLIB_warning("Process already in (other) queue");
+        SIMLIB_warning("Process is already in (other) queue");
         Out();          // if already in queue then remove
     }
     q.Insert(this);     // polymorphic interface
@@ -389,20 +310,21 @@ void Process::Into(Queue & q)
 ////////////////////////////////////////////////////////////////////////////
 /// Process deactivation
 /// To continue the behavior it should be activated again
-/// Warning: memory leak if not activated/deleted
+/// Warning: memory leak if not activated/deleted explicitly (FIXME)
 void Process::Passivate()
 {
     Dprintf(("Process#%lu.Passivate()", id()));
     Entity::Passivate();
     if (!isCurrent())
         return;         // passivated by other process
-    THREAD_INTERRUPT();
+    PROCESS_INTERRUPT();
 }
 
 ////////////////////////////////////////////////////////////////////////////
 /// Terminate the process
-/// if called by current process, self-destruct
-/// free memory of the process
+/// If called by current process, self-destruct.
+/// Remove from queue, unschedule from calendar.
+/// Automatically free the memory of the process.
 void Process::Terminate()
 {
     Dprintf(("Process#%lu.Terminate()", _Ident));
@@ -414,41 +336,76 @@ void Process::Terminate()
     if (!Idle())
         SQS::Get(this);         // remove from calendar
 
-    // End of thread
+    // end of Process
     if (isCurrent()) {          // if currently running process
         _status = _TERMINATED;
-        THREAD_EXIT();          // Jump back to dispatcher
+        PROCESS_EXIT();          // jump back to dispatcher
     }
     else {
         _status = _TERMINATED;
         if (isAllocated())
-            delete this;        // Remove passive process
+            delete this;        // remove passive process
     }
 }
 
 ////////////////////////////////////////////////////////////////////////////
-#define CANARY1 (reinterpret_cast<long>(this)+1) // unaligned value is better
+#define CANARY1 (reinterpret_cast<long>(this)-1) // unaligned value is better
+
+
+/// This is theoretical max stack size for single SIMLIB Process at time of
+/// interruption (should never be too big, because of copying the stack contents)
+/// FIXME: should be checked at Process::Behavior interrupt for more precise error message
+#define MAX_PROCESS_STACK_SIZE  1000000UL
+
+[[noreturn,gnu::noinline]] static void restore_context2(volatile char *) noexcept;
+
+/// This elliminates the need for explicit stack pointer adjustment.
+/// The local array should not be optimized away.
+/// Parameters, return address and local variables will be overwritten.
+/// Never returns.
+[[noreturn,gnu::noinline]] static void restore_context(size_t data_size) noexcept
+{
+    volatile char skip[MAX_PROCESS_STACK_SIZE]; // should never be optimized
+    //skip[MAX_PROCESS_STACK_SIZE-1]='*';
+    if(data_size>MAX_PROCESS_STACK_SIZE)
+        SIMLIB_error("Process stack size limit exceeded");
+    DEBUG_PROCESS(0xB);
+    if(skip > P_StackBase2)
+        SIMLIB_error("Process stack error");
+    restore_context2(skip);
+}
+
+/// This function overwrites stack contents. It never returns, see longjmp.
+[[noreturn,gnu::noinline]] static void restore_context2(volatile char *) noexcept
+{
+            // c) Copy saved stack contents back to stack
+            memcpy((void *) (P_StackBase - P_StackSize), P_Context->stack, P_StackSize);
+            DEBUG_PROCESS(0xC);
+
+            // 4) Restore proces status (registers: SP,IP,...)
+            longjmp(P_Context->status, 1);
+            // ===========================================================
+            // never reach this point - longjmp never returns
+}
 
 /**
  * \fn Process::_Run
  * Process dispatch method
  *
- * The dispatcher starts/reactivates process behavior
+ * The dispatcher starts/reactivates process Behavior() method
  *
  * IMPORTANT notes:
- * - Function contains some non-portable code and should be called
- *   from single place in simulation-control algorithm,
- *   because it is sensitive to stack-frame position
+ * - Function should be called from single place in simulation-control
+ *   algorithm, because it is sensitive to stack-frame position
  *   (it saves/restores the stack contents).
  *
  * This function:
- *  1) marks current position on stack
- *  2) marks current CPU context
- *  3) calls Behavior()
- *  4) after interruption of Behavior() returns
- *  5) if called next time, do 1), 2) and move stack pointer (SP)
- *  6) copy saved stack content back
- *  7) do longjmp() to restore Behavior() execution
+ *  1) saves current context (position on stack and CPU registers)
+ *  2) calls Behavior() like any other function
+ *  3) interruption of Behavior() saves context+ and jumps back
+ *  4) to continue Behavior(), do 1) and move stack pointer away
+ *  5) copy saved stack content back to stack
+ *  6) do longjmp() to restore Behavior() execution
  *
  * @ingroup process
  */
@@ -464,7 +421,7 @@ void Process::_Run() noexcept // no exceptions
         SIMLIB_error(ProcessNotInitialized);
 
     // Mark the stack base address
-    volatile long mylocal = CANARY1;     // should be automatic = on stack
+    volatile long mylocal = CANARY1;     // on stack variable
     // Warning: DO NOT USE ANY OTHER LOCAL VARIABLES in this function!
     P_StackBase = (char*)(&mylocal + 1);
 
@@ -485,13 +442,23 @@ void Process::_Run() noexcept // no exceptions
     //              |   | mylocal2 |
     //              +-> +----------+
     //              |   |          |
-#   define STACK_RESERVED 0x080 // reserved for "red zone" 128B ?
+#   define STACK_RESERVED 0x080 // reserved area for "red zone" 128B ?
     //              |   |          |
     // P_StackBase2 +-> +----------+
+    //                      ...
+    // locals and return address of restore_context2
+    //                      ...
+    //                      ...
+    // SIMLIB stack limit (defined by MAX_PROCESS_STACK_SIZE implementation defined constant)
+    // (limit holds only when Process::Behavior is interrupted and it's context saved)
+    //                      ...
+    //                      ...
+    //                      ...
+    // OS stack size limit (e.g. 8MiB on Linux: ulimit -s)
 
 
 #if EXTRA_DEBUG
-    DEBUG(DBG_THREAD,("| THREAD_STACK_BASE=%016p", P_StackBase));
+    DEBUG(DBG_PROCESS,("| PROCESS_STACK_BASE=%016p", P_StackBase));
     // CHECK if the P_StackBase position is the same in each call
     static char *P_StackBase0=0;
     if(P_StackBase0==0)
@@ -506,9 +473,9 @@ void Process::_Run() noexcept // no exceptions
         // setjmp returned after saving current status
         _status = _RUNNING;
         if (_context == 0) {    // process start
-            DEBUG(DBG_THREAD, ("| --- Process::Behavior() START "));
+            DEBUG(DBG_PROCESS, ("| --- Process::Behavior() START "));
             Behavior();         // run behavior description
-            DEBUG(DBG_THREAD, ("| --- Process::Behavior() END "));
+            DEBUG(DBG_PROCESS, ("| --- Process::Behavior() END "));
             _status = _TERMINATED;
             if(mylocal != CANARY1)
                 SIMLIB_error("Process canary1 died after Behavior() return");
@@ -520,7 +487,7 @@ void Process::_Run() noexcept // no exceptions
                 SQS::Get(this);         // Remove from calendar
             //TODO: if(in any facility) error
         } else { // process was interrupted and has saved context
-            DEBUG(DBG_THREAD, ("| --- Process::Behavior() CONTINUE "));
+            DEBUG(DBG_PROCESS, ("| --- Process::Behavior() CONTINUE "));
             mylocal = 0; // for checking only - previous value should be saved and later restored
             // RESTORE_CONTEXT
             // a) Save local variables to global
@@ -532,20 +499,9 @@ void Process::_Run() noexcept // no exceptions
             // This is very important for next memcpy and longjmp
             // (stack grows down), we reserve some more space
             P_StackBase2 = P_StackBase - P_StackSize - STACK_RESERVED;
-            THREAD_DEBUG(0xB);
 
-            // ==========================================================
-            SET_STACK_PTR(P_StackBase2); // === BEGIN inconsistent state!
-            // Warning: We can not use any local variables here!
-
-            // c) Copy saved stack contents back to stack
-            memcpy((void *) (P_StackBase - P_StackSize), P_Context->stack, P_StackSize);
-            THREAD_DEBUG(0xC);
-
-            // 4) Restore proces status (SP,IP,...)
-            longjmp(P_Context->status, 1); // === END inconsistent state!
-            // ===========================================================
-            // never reach this point - longjmp never returns
+            restore_context(P_StackSize+STACK_RESERVED);
+            // never reach this point - longjmp (called from restore_context) never returns
 
         }
     }
@@ -560,7 +516,7 @@ void Process::_Run() noexcept // no exceptions
             // Store content in global variables back to attributes
             P_Context->size = P_StackSize;
             this->_context = P_Context;
-            DEBUG(DBG_THREAD,("| --- Process::Behavior() INTERRUPT %p.context=%p, size=%d", \
+            DEBUG(DBG_PROCESS,("| --- Process::Behavior() INTERRUPT %p.context=%p, size=%d", \
                                                 this, P_Context, P_StackSize));
             P_Context = 0; // cleaning
         }
@@ -571,7 +527,7 @@ void Process::_Run() noexcept // no exceptions
     //TODO: MOVE to simulation control loop
     if (isTerminated() && isAllocated()) {
         // terminated process on heap
-        DEBUG(DBG_THREAD,("| Process %p ends and is deallocated now",this));
+        DEBUG(DBG_PROCESS,("| Process %p ends and is deallocated now",this));
         delete this;    // destroy process
     }
     // return to simulation control
@@ -581,17 +537,17 @@ void Process::_Run() noexcept // no exceptions
 ////////////////////////////////////////////////////////////////////////////
 #define CANARY2 0xDEADBEEFUL
 /**
- * \fn THREAD_INTERRUPT_f
+ * \fn PROCESS_INTERRUPT_f
  * Special function called from Process::Behavior() directly or indirectly.
  *
  * This function:
- *  1) computes stack content size,
- *  2) allocates memory for stack contents,
- *  3) saves stack contents to allocated memory,
+ *  1) computes stack data size,
+ *  2) allocates memory for stack data,
+ *  3) saves stack data to allocated memory,
  *  4) saves CPU context using setjmp(), and
  *  5) interrupts execution of current function using longjmp()
  *     to process dispatcher code,
- *  == (now run dispatcher and other code)
+ *  == (now runs dispatcher and other code)
  *  6) continues execution after longjmp from dispatcher.
  *  7) frees memory allocated at 2)
  *
@@ -600,33 +556,34 @@ void Process::_Run() noexcept // no exceptions
  *
  * @ingroup process
  */
-static void THREAD_INTERRUPT_f()
+[[gnu::noinline]] static void PROCESS_INTERRUPT_f()
 {
     // SAVE THE STACK STATE of the thread
 
-    // 1) compute stack context size  (from P_StackBase to local variable)
-    volatile unsigned mylocal2 = CANARY2;       // the only on-stack variable
+    // 1) compute stack context size  (from P_StackBase to next variable)
+    volatile unsigned mylocal2 = CANARY2;       // on-stack variable
+    // Warning: DO NOT USE ANY OTHER LOCAL VARIABLES in this function!
     P_StackSize = (size_t) (P_StackBase - (char *) (&mylocal2));
-    THREAD_DEBUG(1);
+    DEBUG_PROCESS(1);
 
-    // 2) allocate memory for stack contents
+    // 2) allocate memory for current context
     ALLOC_CONTEXT(P_StackSize);
 
     // 3) save stack data (stack grows DOWN)
     memcpy(P_Context->stack, (P_StackBase - P_StackSize), P_StackSize);
 
-    mylocal2 = 0;       // previous value should be saved now and restored later
+    mylocal2 = 0;       // previous value is saved (will be restored later)
 
     // STACK CONTENTS SAVED
 
-    // 4) mark CPU context
-    THREAD_DEBUG(4);
+    // 4) save CPU context (see man setjmp for details)
+    DEBUG_PROCESS(4);
     if (!setjmp(P_Context->status)) {
         ////////////////////////////////////////////////////////////////////
         // 5) Interrupt the execution of Behavior()
-        THREAD_DEBUG(5);
+        DEBUG_PROCESS(5);
         longjmp(P_DispatcherStatusBuffer, 1);     // --> longjmp back to dispatcher
-        // never return here
+        // longjmp never returns
     }
 
     // Check canary on stack after restore
@@ -635,13 +592,13 @@ static void THREAD_INTERRUPT_f()
 
     ////////////////////////////////////////////////////////////////////////
     // 6) Continue execution after longjmp from dispatcher
-    // Data were restored on stack, longjmp restored SP
-    THREAD_DEBUG(6);
+    // On stack data and CPU context were restored by dispatcher
+    DEBUG_PROCESS(6);
 
-    // 7) free memory of stack contents
+    // 7) free memory of already restored context
     FREE_CONTEXT();
-    THREAD_DEBUG(7);
-    // return and continue in Process::Behavior() execution
+    DEBUG_PROCESS(7);
+    // return and continue Process::Behavior() execution
 }
 
 } // namespace
