@@ -5,10 +5,10 @@
 
 #ifndef __SIMLIB__
 
-#define __SIMLIB__        0x0308 // version of SIMLIB (numerical)
-#define SIMLIB_VERSION    "3.08" // version of SIMLIB (string)
+#define __SIMLIB__        0x0309 // version of SIMLIB (numerical)
+#define SIMLIB_VERSION    "3.09" // version of SIMLIB (string)
 
-#define SIMLIB_COPYRIGHT  "(c) Petr Peringer, 1991-2021"
+#define SIMLIB_COPYRIGHT  "(c) Petr Peringer, 1991-2022"
 // -------------------------------------------------------------------------
 // SIMLIB/C++ (SIMulation LIBrary for C++)
 // -------------------------------------------------------------------------
@@ -87,7 +87,7 @@ const char *SIMLIB_version_string();    // get SIMLIB version string
 # define SIMLIB_COMPILER  "GNU C++"
 
 #else
-# error  "SIMLIB is not implemented for this system/compiler"
+# warning  "SIMLIB is not tested on this system/compiler"
 #endif
 
 ////////////////////////////////////////////////////////////////////////////
@@ -435,7 +435,7 @@ inline void Passivate(Entity *e) { e->Passivate(); }  //!< passivate entity e
 //! Abstract base class for all simulation processes
 //! @ingroup process
 //! Process behavior is specified by Behavior method and can be interrupted
-//! during execution by e.g. Wait method call (it is coroutine)
+//! during execution by e.g. Wait method call (it is a coroutine)
 //! \ingroup simlib
 class Process : public Entity {
   void * _context;                      //!< process context pointer
